#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Ciclope module for tetrahedra Finite Element model generation
"""

import logging
import numpy as np
import meshio
import mcubes

def shell_mesh(bwimage, method='pymcubes', voxelsize=[1., 1., 1.], max_facet_distance=0.0, max_cell_circumradius=0.0):
    """Generate outer shell mesh of triangles from binary volume data.
    The mesh is generated using the PyMCubes module and the smooth function contained in it:
    https://github.com/pmneila/PyMCubes

    Alternatively, the marching cube algorithm from the scikit-image python module can be used:
    https://scikit-image.org/docs/dev/api/skimage.measure.html?highlight=marching#skimage.measure.marching_cubes


    Parameters
    ----------
    bwimage
        Binary image.
    method : str
        'pymcubes': PyMCubes module.
        'marching_cubes': scikit-image's marching cube algorithm.
        'pygalmesh': pygalmesh module (CGAL).
    voxelsize : float
        Image voxelsize.
    max_facet_distance : float
        CGAL parameter.
    max_cell_circumradius : float
        CGAL parameter.

    Returns
    -------
    vertices
        Mesh vertices.
    triangles
        Mesh triangles.
    shellmesh : meshio
        Mesh data.
    """

    if method == 'pymcubes':
        # (the 0-levelset of the output of mcubes.smooth is the smoothed version of the 0.5-levelset of the binary array.
        smoothed_L = mcubes.smooth(bwimage)
        # Extract the 0-levelset
        vertices, triangles = mcubes.marching_cubes(np.transpose(smoothed_L, [2, 1, 0]), 0)
        shellmesh = None

    elif method == 'marching_cubes':
        from skimage.measure import marching_cubes
        vertices, triangles, tmp, tmp2 = marching_cubes(np.transpose(bwimage, [2, 1, 0]), level=None, step_size=1)
        shellmesh = None

    elif method == 'pygalmesh':
        shellmesh = cgal_mesh(bwimage, voxelsize, 'triangle', max_facet_distance, max_cell_circumradius)
        vertices = None
        triangles = None

    else:
        raise IOError('{0} method unknown.', format(method))

    # get SOUTH cap vertices
    # import pygmsh
    # vertices_capS = vertices[vertices[:, 0] == vertices[:, 0].min()]
    # with pygmsh.geo.Geometry() as geom:
    #     geom.add_polygon(vertices_capS[:,1:3], mesh_size=1)
    #     mesh_capS = geom.generate_mesh()
    #
    # import pygalmesh
    # mesh_capS = pygalmesh.generate_2d(vertices_capS[:,1:3],max_edge_size=0.2,num_lloyd_steps=10)

    return vertices, triangles, shellmesh

def check_cgal_params(max_facet_distance, max_cell_circumradius, voxelsize):
    """Check CGAL mesher parameters.
    # https://github.com/nschloe/pygalmesh#volume-meshes-from-surface-meshes

    Parameters
    ----------
    max_facet_distance
    max_cell_circumradius

    voxelsize : float
        Image voxel size.

    Returns
    -------
    max_facet_distance : float
    max_cell_circumradius : float
    """

    if len(voxelsize) > 1:
        voxelsize = voxelsize[0]

    if max_facet_distance is None:
        max_facet_distance = 1 * voxelsize

    if max_cell_circumradius is None:
        max_cell_circumradius = 5 * voxelsize

    return max_facet_distance, max_cell_circumradius

def cgal_mesh(bwimage, voxelsize, meshtype='both', max_facet_distance=0.0, max_cell_circumradius=0.0):
    """Generate mesh of from binary volume data using CGAL.
    The mesh is generated using the PyGalmesh module. For more info visit: https://github.com/nschloe/pygalmesh#volume-meshes-from-surface-meshes
    The pygalmesh.generate_from_array method returns a mesh containing both a cells set of tetrahedra (volume mesh)
    and a cells set of triangles (shell mesh). The parameter 'meshtype' is used to control which type of mesh is returned.

    Parameters
    ----------
    bwimage
        Binary image.
    voxelsize : float
        Image voxelsize.
    meshtype : str
        'triangle': Outer mesh (shell) of triangles.
        'tetra': Volume mesh of tetrahedra.
        'both': Both shell and volume cells sets.
    max_facet_distance : float
        CGAL parameter.
    max_cell_circumradius : float
        CGAL parameter.

    Returns
    -------
    mesh : meshio
        Mesh data.

    """

    import pygalmesh
    # generate mesh
    mesh = pygalmesh.generate_from_array(np.transpose(bwimage, [2, 1, 0]).astype('uint8'), tuple(voxelsize), max_facet_distance=max_facet_distance, max_cell_circumradius=max_cell_circumradius)

    # check mesh fields and get index for triangle and tetra cells
    triangle_id = None
    tetra_id = None

    for id in range(0, len(mesh.cells)):
        if mesh.cells[id].type == 'triangle':
            triangle_id = id
        elif mesh.cells[id].type == 'tetra':
            tetra_id = id

    # pygalmesh returns two cells entities ('triangle' and 'tetra'). Select one or both for return
    if len(mesh.cells) > 1:
        if meshtype == 'triangle':
            if triangle_id is None:
                raise Exception('pygalmesh did not return a triangle mesh.')
            else:
                logging.info("Removing tetra cells.")
                mesh.cells = [mesh.cells[triangle_id]]
                # this is prone to errors..
                mesh.cell_data['medit:ref'] = mesh.cell_data['medit:ref'][triangle_id]
        elif meshtype == 'tetra':
            if tetra_id is None:
                raise Exception('pygalmesh did not return a tetra mesh.')
            else:
                logging.info("Removing triangle cells.")
                mesh.cells = [mesh.cells[tetra_id]]
                # this is prone to errors..
                mesh.cell_data['medit:ref'] = mesh.cell_data['medit:ref'][tetra_id]
        elif meshtype == 'both':
            logging.info("Both triangle and tetra cells kept.")
        else:
            raise IOError('{0} method type.', format(meshtype))

    return mesh

def mesh2tetrafe(meshdata, templatefile, fileout, keywords=['NSET', 'ELSET'], float_fmt='.6e', verbose=False):
    """Generate ABAQUS tetrahedra Finite Element (FE) input file from 3D mesh. The output can be solved using ABAQUS or CalculiX.

    Boundary conditions (BCs), simulation steps and associated output requests are defined in a separate template file.
    See the templates contained in the folder "input_templates" for examples.

    Parameters
    ----------
    meshdata : meshio
        Mesh data.
    templatefile : str
        Analysis template file.
    fileout : str
        Output .INP file.
    keywords : str
        SUPPORTED ABAQUS KEYWORDS:

        For a list of all Abaqus keywords and their description visit:
        https://abaqus-docs.mit.edu/2017/English/SIMACAECAERefMap/simacae-c-gen-kwbrowser.htm#simacae-c-gen-kwbrowser__simacae-gen-xsl-U

        * 'NSET':
                Create boundary node sets. (Default = ON)
                If 'NSET' is specified, the following node sets are created:
                  - NODES_X0: Nodes on WEST (X-) surface of 3D model.
                  - NODES_X1: Nodes on EAST (X+) surface of 3D model.
                  - NODES_Y0: Nodes on SOUTH (Y-) surface of 3D model.
                  - NODES_Y1: Nodes on NORTH (Y+) surface of 3D model.
                  - NODES_Z0: Nodes on BOTTOM (Z-) surface of 3D model.
                  - NODES_Z1: Nodes on TOP (Z+) surface of 3D model.
                  - NODES_X0Y0Z0: 2 nodes on (0,0,0) model corner.
                  - NODES_X0Y0Z1: 2 nodes on (0,0,1) model corner.
                These node sets are available for boundary conditions definition.
        * 'ELSET':
                Create boundary element sets. (Default = ON)
                If 'ELSET' is specified, the following element sets are created:
                  - ELEMS_X0: Elements of WEST (X-) surface of 3D model.
                  - ELEMS_X1: Elements of EAST (X+) surface of 3D model.
                  - ELEMS_Y0: Elements of SOUTH (Y-) surface of 3D model.
                  - ELEMS_Y1: Elements of NORTH (Y+) surface of 3D model.
                  - ELEMS_Z0: Elements of BOTTOM (Z-) surface of 3D model.
                  - ELEMS_Z1: Elements of TOP (Z+) surface of 3D model.
    float_fmt : float
        Precision for Abaqus input file writing.
    verbose : bool
        Verbose output.
    """

    # verbose output
    if verbose is True:
        logging.basicConfig(level=logging.INFO)

    logging.info('Converting mesh with fields:\n')
    logging.info(vars(meshdata))

    if 'NSET' in keywords:
        # find model boundaries
        model_coors_max = np.amax(meshdata.points, 0)
        model_coors_min = np.amin(meshdata.points, 0)

        # volume extent in x, y and z
        extent = model_coors_max-model_coors_min

        # add dictionary of boundary point sets
        meshdata.point_sets = {
            'NODES_Y1': np.where(meshdata.points[:, 1] == model_coors_max[1])[0],
            'NODES_Y0': np.where(meshdata.points[:, 1] == model_coors_min[1])[0],
            'NODES_X1': np.where(meshdata.points[:, 0] == model_coors_max[0])[0],
            'NODES_X0': np.where(meshdata.points[:, 0] == model_coors_min[0])[0],
            'NODES_Z1': np.where(meshdata.points[:, 2] >= (model_coors_max[2]-0.01*abs(extent[2])))[0],
            'NODES_Z0': np.where(meshdata.points[:, 2] <= (model_coors_min[2]+0.01*abs(extent[2])))[0]
        }

    meshdata.point_sets['NODES_X0Y0Z0'] = meshdata.point_sets['NODES_Z0'][0:2]
    meshdata.point_sets['NODES_X0Y0Z1'] = meshdata.point_sets['NODES_Z1'][0:2]

    if 'ELSET' in keywords:
        # dictionary of boundary element sets
        elset = {
            'ELEMS_Y1': [],
            'ELEMS_Y0': [],
            'ELEMS_X0': [],
            'ELEMS_X1': [],
            'ELEMS_Z1': [],
            'ELEMS_Z0': []
        }

        # find boundary cells
        ELEMS_X1 = np.array([]).astype('int')
        ELEMS_X0 = np.array([]).astype('int')
        ELEMS_Y0 = np.array([]).astype('int')
        ELEMS_Y1 = np.array([]).astype('int')
        ELEMS_Z1 = np.array([]).astype('int')
        ELEMS_Z0 = np.array([]).astype('int')

        for node_e in meshdata.point_sets['NODES_X1']:
            ELEMS_X1 = np.append(ELEMS_X1, np.where(np.any(meshdata.cells[0][1] == node_e, axis=1)))

        for node_w in meshdata.point_sets['NODES_X0']:
            ELEMS_X0 = np.append(ELEMS_X0, np.where(np.any(meshdata.cells[0][1] == node_w, axis=1)))

        for node_s in meshdata.point_sets['NODES_Y0']:
            ELEMS_Y0 = np.append(ELEMS_Y0, np.where(np.any(meshdata.cells[0][1] == node_s, axis=1)))

        for node_n in meshdata.point_sets['NODES_Y1']:
            ELEMS_Y1 = np.append(ELEMS_Y1, np.where(np.any(meshdata.cells[0][1] == node_n, axis=1)))

        for node_t in meshdata.point_sets['NODES_Z1']:
            ELEMS_Z1 = np.append(ELEMS_Z1, np.where(np.any(meshdata.cells[0][1] == node_t, axis=1)))

        for node_b in meshdata.point_sets['NODES_Z0']:
            ELEMS_Z0 = np.append(ELEMS_Z0, np.where(np.any(meshdata.cells[0][1] == node_b, axis=1)))

        meshdata.cell_sets = {'SETALL': [np.arange(0, len(meshdata.cells[0][1]))]}

        # if cell data exists create element sets based on it
        if hasattr(meshdata, 'cell_data'):

            # find unique cell data
            cell_data_unique = np.unique(meshdata.cell_data['medit:ref'])

            # for each unique cell scalar find indexes of the corresponding cells and create element set of them
            for val in cell_data_unique:
                meshdata.cell_sets['SET' + str(val)] = [np.where(meshdata.cell_data['medit:ref'] == val)[0]]
                # print(val)

        if not ELEMS_Y1.size == 0:
            meshdata.cell_sets['ELEMS_Y1'] = [np.unique(ELEMS_Y1)]

        if not ELEMS_Y0.size == 0:
            meshdata.cell_sets['ELEMS_Y0'] = [np.unique(ELEMS_Y0)]

        if not ELEMS_X0.size == 0:
            meshdata.cell_sets['ELEMS_X0'] = [np.unique(ELEMS_X0)]

        if not ELEMS_X1.size == 0:
            meshdata.cell_sets['ELEMS_X1'] = [np.unique(ELEMS_X1)]

        if not ELEMS_Z1.size == 0:
            meshdata.cell_sets['ELEMS_Z1'] = [np.unique(ELEMS_Z1)]

        if not ELEMS_Z0.size == 0:
            meshdata.cell_sets['ELEMS_Z0'] = [np.unique(ELEMS_Z0)]

    else:
        meshdata.cell_sets = {'SETALL': [np.arange(0, len(meshdata.cells[0][1]))]}

    # write Abaqus mesh using meshio
    meshio.abaqus.write(fileout, meshdata, float_fmt)

    # Add analysis definition to the end of the Abaqus INP from template
    # Open ABAQUS *.INP output file in Append mode
    INP = open(fileout, 'a')

    # copy line by line info on model solution and boundary conditions from Abaqus template file
    try:
        BCfile = open(templatefile, 'r')
    except IOError('Abaqus template file {} not found.'.format(templatefile)):
        exit(1)
    logging.info('Reading Abaqus template file {}'.format(templatefile))

    for line in BCfile.readlines():
        # copy line to output Abaqus file
        INP.write('{}'.format(line))

    BCfile.close()
    INP.close()
    logging.info('Data written to file {}'.format(fileout))




import ciclope.core.voxelFE
from ciclope.utils.recon_utils import read_tiff_stack
from ciclope.utils.preprocess import remove_unconnected
import ciclope
import numpy as np
import unittest

test_data_file = 'trab/slice_00.tif'

class run_tests(unittest.TestCase):
    def test_remove_unconnected(self):
        # read and segment test dataset
        pippo_BW = read_tiff_stack(test_data_file) > 100
        pippo_BW_remove = remove_unconnected(pippo_BW)
        self.assertEqual((np.sum(pippo_BW) - np.sum(pippo_BW_remove)), 4)

    def test_tetraFE_shell_mesh_pymcubes(self):
        # (pymcubes method)
        vert, tri, mesh = ciclope.tetraFE.shell_mesh(remove_unconnected((read_tiff_stack(test_data_file)>100)), method='pymcubes', voxelsize=[0.12, 0.12, 0.12])
        # check if 3D shell mesh fields (vertices, triangles) were produced
        self.assertEqual(vert.shape[1], 3)
        self.assertEqual(tri.shape[1], 3)

    def test_tetraFE_shell_mesh_pygalmesh(self):
        # (pygalmesh method)
        vert, tri, mesh = ciclope.tetraFE.shell_mesh(remove_unconnected((read_tiff_stack(test_data_file)>100)), method='pygalmesh', voxelsize=[0.12, 0.12, 0.12])
        self.assertNotEqual(len(mesh.points), 0)
        self.assertEqual(mesh.cells[0][0], 'triangle')
        self.assertNotEqual(len(mesh.cells[0][1]), 0)

    def test_tetraFE_cgal_mesh(self):
        mesh = ciclope.tetraFE.cgal_mesh(remove_unconnected((read_tiff_stack(test_data_file)>100)), [0.12, 0.12, 0.12], meshtype='tetra')
        # check that points and cells fields are not empty
        self.assertNotEqual(len(mesh.points), 0)
        self.assertEqual(mesh.cells[0][0], 'tetra')
        self.assertNotEqual(len(mesh.cells[0][1]), 0)

    def test_mesh2tetrafe(self):
        ciclope.tetraFE.mesh2tetrafe(ciclope.tetraFE.cgal_mesh(remove_unconnected((read_tiff_stack(test_data_file)>100)), [0.12, 0.12, 0.12], meshtype='tetra'), 'test_input.inp', 'foo.inp')
        # read and check input CALCULIX file: check existence of essential fields
        with open("foo.inp", 'r') as f:
            lines = f.readlines()
            node = False
            element = False
            elset = False
            elset_z1 = False
            nset_z0 = False
            step = False
            bc = False

            for line in lines:
                if line.find('*NODE') != -1:
                    node = True
                if line.find('*ELEMENT, TYPE=C3D4') != -1:
                    element = True
                if line.find('*ELSET, ELSET=SETALL') != -1:
                    elset = True
                if line.find('*ELSET, ELSET=ELEMS_Z1') != -1:
                    elset_z1 = True
                if line.find('*NSET, NSET=NODES_Z0') != -1:
                    nset_z0 = True
                if line.find('*STEP') != -1:
                    step = True
                if line.find('NODES_T, 3, 3, -0.002') != -1:
                    bc = True

        self.assertEqual(node, True)
        self.assertEqual(element, True)
        self.assertEqual(elset, True)
        self.assertEqual(elset_z1, True)
        self.assertEqual(nset_z0, True)
        self.assertEqual(step, True)
        self.assertEqual(bc, True)

    def test_voxelFE_vol2ugrid(self):
        pippo_mesh = ciclope.core.voxelFE.vol2ugrid(read_tiff_stack(test_data_file), [0.12,0.12,0.12], GVmin=100)
        self.assertEqual(pippo_mesh.cells[0].type, "hexahedron")
        self.assertEqual(pippo_mesh.cells[0].data.shape, (209,8))
        self.assertEqual(pippo_mesh.cells[0].data.shape, (209,8))
        self.assertEqual(pippo_mesh.points.shape, (1331,3))
        self.assertEqual(pippo_mesh.point_sets['NODES_X0Y0Z0'], [18,19])
        self.assertEqual((pippo_mesh.points[100] == np.array([0.12, 1.08, 0])).all(), True)

    def test_mesh2voxelFE(self):
        ciclope.voxelFE.mesh2voxelfe(ciclope.core.voxelFE.vol2ugrid(read_tiff_stack(test_data_file), [0.12,0.12,0.12], GVmin=100), 'test_input.inp', 'foo.inp')
        # read and check input CALCULIX file: check existence of essential fields
        with open("foo.inp", 'r') as f:
            lines = f.readlines()
            node = False
            element = False
            elset_z1 = False
            nset_z0 = False
            step = False
            bc = False

            for line in lines:
                if line.find('*NODE') != -1:
                    node = True
                if line.find('*ELEMENT, TYPE=C3D8, ELSET=SET255') != -1:
                    element = True
                if line.find('*ELSET, ELSET=CELLS_Z1') != -1:
                    elset_z1 = True
                if line.find('*NSET, NSET=NODES_Z0') != -1:
                    nset_z0 = True
                if line.find('*STEP') != -1:
                    step = True
                if line.find('NODES_T, 3, 3, -0.002') != -1:
                    bc = True

        # read and check input CALCULIX file: read two specific file lines
        with open("foo.inp", 'r') as f:
            line_numbers = [99, 860]
            lines = []
            for i, line in enumerate(f):
                if i in line_numbers:
                    lines.append(line.strip())

        self.assertEqual(node, True)
        self.assertEqual(element, True)
        self.assertEqual(elset_z1, True)
        self.assertEqual(nset_z0, True)
        self.assertEqual(step, True)
        self.assertEqual(bc, True)

        # assert one node and one node_set line
        self.assertEqual(list(map(float, lines[0].split(","))), [359.0, 0.84, 1.2, 0.24])
        self.assertEqual(list(map(int, lines[1].strip(',').split(','))), [37, 64, 92, 126])

if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Ciclope module for voxel Finite Element model generation
"""

import os
import logging
import numpy as np
from datetime import datetime

def vol2ugrid(voldata, voxelsize=[1, 1, 1], GVmin=0, refnodes=False, verbose=False):
    """Generate unstructured grid mesh from 3D volume data.

    Parameters
    ----------
    voldata : ndarray
        3D voxel data.
    voxelsize : float
        3D model voxelsize.
    matpropbits : int
        Bit depth for material mapping.
    GVmin
        Minimum Grey Value considered for meshing. By default, GVmin=0: all zeroes are considered background.
    refnodes : bool
        Return dictionary of reference nodes on the model boundaries. Even if this option is not activated, the returned
        mesh will contain the following nodes and elements sets:
            - NODES_X0: Nodes on WEST (X-) surface of 3D model.
            - NODES_X1: Nodes on EAST (X+) surface of 3D model.
            - NODES_Y0: Nodes on SOUTH (Y-) surface of 3D model.
            - NODES_Y1: Nodes on NORTH (Y+) surface of 3D model.
            - NODES_Z0: Nodes on BOTTOM (Z-) surface of 3D model.
            - NODES_Z1: Nodes on TOP (Z+) surface of 3D model.
            - NODES_X0Y0Z0: 2 nodes on (0,0,0) model corner.
            - NODES_X0Y0Z1: 2 nodes on (0,0,1) model corner.

            - ELEMS_X0: Elements of WEST (X-) surface of 3D model.
            - ELEMS_X1: Elements of EAST (X+) surface of 3D model.
            - ELEMS_Y0: Elements of SOUTH (Y-) surface of 3D model.
            - ELEMS_Y1: Elements of NORTH (Y+) surface of 3D model.
            - ELEMS_Z0: Elements of BOTTOM (Z-) surface of 3D model.
            - ELEMS_Z1: Elements of TOP (Z+) surface of 3D model.
    verbose : bool
        Activate verbose output.

    Returns
    -------
    mesh : meshio
        Unstructured grid mesh.
    refnodes : list
        centroids on the model boundaries (X0, X1, Y0, Y1, Z0, Z1)
    """

    # verbose output
    if verbose:
        logging.basicConfig(level=logging.INFO)

    # get 3D dataset shape
    data_shape = voldata.shape

    # voxelsize list if only one value is given
    if type(voxelsize) is not list and type(voxelsize) is not np.ndarray:
        voxelsize = [voxelsize, voxelsize, voxelsize]
    if len(voxelsize) == 1:
        voxelsize = [voxelsize[0], voxelsize[0], voxelsize[0]]

    # variables initialization ##################################################
    # nodes arrays
    nodes = []
    existing_nodes = np.full((data_shape[0]+1)*(data_shape[1]+1)*(data_shape[2]+1), False)

    # cell nodes and cell grey value arrays
    cells = []
    cell_GV = []

    # dictionary of boundary node sets
    nodes_set = {
        'NODES_Y1': [],
        'NODES_Y0': [],
        'NODES_X0': [],
        'NODES_X1': [],
        'NODES_Z1': [],
        'NODES_Z0': [],
        'NODES_X0Y0Z0': [],
        'NODES_X0Y0Z1': []
    }

    # reference nodes coordinates (one refnode per boundary)
    refnode_X0 = np.zeros(3)
    refnode_X1 = np.zeros(3)
    refnode_Y0 = np.zeros(3)
    refnode_Y1 = np.zeros(3)
    refnode_Z0 = np.zeros(3)
    refnode_Z1 = np.zeros(3)

    # dictionary of boundary cell sets
    cells_set = {
        'CELLS_Y1': [],
        'CELLS_Y0': [],
        'CELLS_X0': [],
        'CELLS_X1': [],
        'CELLS_Z1': [],
        'CELLS_Z0': []
    }

    # cell data ##################################################
    row_nodes = data_shape[1] + 1  # n nodes along y
    slice_nodes = (data_shape[2] + 1) * (data_shape[1] + 1)  # n nodes in 1 slice
    cell_i = 0

    logging.info('Calculating cell array')
    for slice in range(data_shape[0]):
        for row in range(data_shape[1]):
            for col in range(data_shape[2]):

                # get cell GV
                GV = voldata[(slice, row, col)]

                if GV > GVmin:
                    cell_i = cell_i + 1
                    # cell nodes indexes
                    # See eight-node brick cell (C3D8 and F3D8) node convention (Par. 6.2.1) at: http://www.dhondt.de/ccx_2.15.pdf
                    cell_nodes = [slice_nodes * slice + row_nodes * row + col,
                                  slice_nodes * slice + row_nodes * row + col + 1,
                                  slice_nodes * slice + row_nodes * (row + 1) + col + 1,
                                  slice_nodes * slice + row_nodes * (row + 1) + col,
                                  slice_nodes * (slice + 1) + row_nodes * row + col,
                                  slice_nodes * (slice + 1) + row_nodes * row + col + 1,
                                  slice_nodes * (slice + 1) + row_nodes * (row + 1) + col + 1,
                                  slice_nodes * (slice + 1) + row_nodes * (row + 1) + col]

                    # append to cell list
                    cells.append(cell_nodes)

                    # append to cell_GV list
                    cell_GV.append(GV)

                    # dictionary of cells belonging to boundary sets
                    if slice == 0:
                        cells_set['CELLS_Z0'].append(cell_i)
                    if slice == data_shape[0] - 1:
                        cells_set['CELLS_Z1'].append(cell_i)
                    if col == 0:
                        cells_set['CELLS_X0'].append(cell_i)
                    if col == data_shape[2] - 1:
                        cells_set['CELLS_X1'].append(cell_i)
                    if row == 0:
                        cells_set['CELLS_Y0'].append(cell_i)
                    if row == data_shape[1] - 1:
                        cells_set['CELLS_Y1'].append(cell_i)

                    # store if the cell nodes exist
                    for i in cell_nodes:
                        existing_nodes[i] = True

    # node coordinates and boundary node sets
    logging.info('Detecting node coordinates and boundary nodes')
    node_i = 0
    for slice in range(data_shape[0] + 1):
        for row in range(data_shape[1] + 1):
            for col in range(data_shape[2] + 1):
                # store node coordinates
                node_coors = [voxelsize[0] * col, voxelsize[1] * row, voxelsize[2] * slice]
                nodes.append(node_coors)

                if existing_nodes[node_i]:
                    # compose dictionary of boundary node sets
                    if slice == 0:
                        nodes_set['NODES_Z0'].append(node_i)
                        refnode_Z0 += node_coors
                    if slice == data_shape[0]:
                        nodes_set['NODES_Z1'].append(node_i)
                        refnode_Z1 += node_coors
                    if col == 0:
                        nodes_set['NODES_X0'].append(node_i)
                        refnode_Y0 += node_coors
                    if col == data_shape[2]:
                        nodes_set['NODES_X1'].append(node_i)
                        refnode_Y1 += node_coors
                    if row == 0:
                        nodes_set['NODES_Y0'].append(node_i)
                        refnode_X0 += node_coors
                    if row == data_shape[1]:
                        nodes_set['NODES_Y1'].append(node_i)
                        refnode_X1 += node_coors

                node_i = node_i + 1

    nodes_set['NODES_X0Y0Z0'] = nodes_set['NODES_Z0'][0:2]
    nodes_set['NODES_X0Y0Z1'] = nodes_set['NODES_Z1'][0:2]

    # reference nodes are the barycenters of boundary node sets
    refnode_X0 /= len(nodes_set['NODES_X0'])
    refnode_X1 /= len(nodes_set['NODES_X1'])
    refnode_Y0 /= len(nodes_set['NODES_Y0'])
    refnode_Y1 /= len(nodes_set['NODES_Y1'])
    refnode_Z0 /= len(nodes_set['NODES_Z0'])
    refnode_Z1 /= len(nodes_set['NODES_Z1'])

    refnodes_dict = {
                   "X0": refnode_X0,
                   "X1": refnode_X1,
                   "Y0": refnode_Y0,
                   "Y1": refnode_Y1,
                   "Z0": refnode_Z0,
                   "Z1": refnode_Z1,
    }

    # generate meshio object
    import meshio

    cells_dict = [("hexahedron", cells)]
    # mesh = meshio.Mesh(nodes, cells_dict, cell_data={"GV": [cell_GV]})
    mesh = meshio.Mesh(nodes, cells_dict, cell_data={"GV": [cell_GV]}, point_sets=nodes_set, cell_sets=cells_set)
    # mesh = meshio.Mesh(nodes, cells_dict, cell_data={"GV": [np.array(cell_GV, dtype='float')]})
    # mesh.write("foo.vtk")

    logging.info('Generated the following mesh with {0} nodes and {1} elements:'.format(len(mesh.points), len(mesh.cells[0])))
    logging.info(mesh)

    if refnodes:
        return mesh, refnodes_dict
    else:
        return mesh

def mesh2voxelfe(mesh, templatefile, fileout, matprop=None, keywords=['NSET', 'ELSET'], eltype='C3D8', matpropbits=8, refnode=None, verbose=False):
    """Generate ABAQUS voxel Finite Element (FE) input file from 3D Unstructured Grid mesh data.
    The file written is an input file (.INP) in ABAQUS syntax that can be solved using ABAQUS or CALCULIX.
    The user can define a material mapping strategy for the conversion of local GVs to local material properties in the FE model.
    Material mapping laws are defined in separate template file(s) (see "prop.inp" and "property_temp_bone.inp" for examples).
    Boundary conditions, analysis type and output requests are defined in a separate template file (see "tmp.inp" for an example).
    Info on analysis definition at: https://abaqus-docs.mit.edu/2017/English/SIMACAECAERefMap/simacae-m-Sim-sb.htm#simacae-m-Sim-sb

    Parameters
    ----------
    mesh : meshio
        Unstructured grid mesh.
    templatefile : str
        Analysis template file.
    fileout : str
        Output .INP file.
    matprop : dict
        Dictionary of material properties for material property mapping:
        matprop = {
            "file": ["prop.inp", "property_temp_bone.inp", ...],
            "range": [[250, 255], [0, 250], ...],
        }
    keywords : str
        SUPPORTED ABAQUS KEYWORDS:

        For a list of all Abaqus keywords and their description visit:
        https://abaqus-docs.mit.edu/2017/English/SIMACAECAERefMap/simacae-c-gen-kwbrowser.htm#simacae-c-gen-kwbrowser__simacae-gen-xsl-U

        * 'NSET':
                Create boundary node sets. (Default = ON)
                If 'NSET' is specified, the following node sets are created:
                  - NODES_X0: Nodes on WEST (X-) surface of 3D model.
                  - NODES_X1: Nodes on EAST (X+) surface of 3D model.
                  - NODES_Y0: Nodes on SOUTH (Y-) surface of 3D model.
                  - NODES_Y1: Nodes on NORTH (Y+) surface of 3D model.
                  - NODES_Z0: Nodes on BOTTOM (Z-) surface of 3D model.
                  - NODES_Z1: Nodes on TOP (Z+) surface of 3D model.
                  - NODES_X0Y0Z0: 2 nodes on (0,0,0) model corner.
                  - NODES_X0Y0Z1: 2 nodes on (0,0,1) model corner.
                These node sets are available for boundary conditions definition.
        * 'ELSET':
                Create boundary element sets. (Default = ON)
                If 'ELSET' is specified, the following element sets are created:
                  - ELEMS_X0: Elements of WEST (X-) surface of 3D model.
                  - ELEMS_X1: Elements of EAST (X+) surface of 3D model.
                  - ELEMS_Y0: Elements of SOUTH (Y-) surface of 3D model.
                  - ELEMS_Y1: Elements of NORTH (Y+) surface of 3D model.
                  - ELEMS_Z0: Elements of BOTTOM (Z-) surface of 3D model.
                  - ELEMS_Z1: Elements of TOP (Z+) surface of 3D model.
        * 'PROPERTY':
                Define an external material mapping law from template file. (Default = None)
                Use in combination with 'matprop' dictionary of material property files and corresponding GV ranges for the material mapping.
    eltype : str
        FE element type. The default is eight-node brick element (C3D8 and F3D8). See CalculiX node convention (Par. 6.2.1) at: http://www.dhondt.de/ccx_2.15.pdf
    matpropbits : int
        Bit depth for material mapping.
    refnode
        Reference node coordinates [REF_NODE_x, REF_NODE_y, REF_NODE_z] for kinematic coupling.
    verbose : bool
        Activate verbose output.
    """

    # verbose output
    if verbose:
        logging.basicConfig(level=logging.INFO)

    # variables initialization ##################################################
    # dictionaries (two) of GVs for each user-defined material property
    prop_GV = {}
    prop_GV2 = {}

    # dictionary of the GVs existing in the input model
    # existing_GV = []
    cell_GV = []
    GVmin = 0
    GVmax = 2 ** matpropbits - 1

    # check input model cell data
    try:
        mesh.cell_data
    except NameError:
        logging.warning('Mesh does not contain cell data. Material property mapping disabled.')
        binary_model = True
    else:
        binary_model = False
        keys = list(mesh.cell_data)
        cell_GV = mesh.cell_data.get(keys[0])
        if type(cell_GV) is list:
            cell_GV = cell_GV[0]
        # cell_GV = mesh.cell_data.get(keys[0])[0]
        logging.info('Found cell_data: {keyname}. cell_data range: {0} - {1}.'.format(min(cell_GV), max(cell_GV), keyname=keys[0]))
        if (min(cell_GV) == 0) & (max(cell_GV) == 1) & (issubclass(cell_GV.dtype.type, np.integer)):
            logging.warning('cell_data is binary. Material property mapping disabled.')
            binary_model = True

    # existing grey values
    existing_GV = np.unique(cell_GV)
    if max(existing_GV) > GVmax:
        GVmax = max(existing_GV)

    # initialize dictionaries of user material properties if given
    if binary_model:
        # GV range for binary dataset
        prop_GV[0] = (1, 1)
    else:
        if matprop is not None:
            # check matprop dictionary input
            if not "range" in matprop and len(matprop["file"]) > 1:
                raise IOError('Incorrect material property definition: GV ranges required when using multiple property files.')
            elif not "range" in matprop and len(matprop["file"]) == 1:
                # assign all GVs above zero to the given material property
                matprop["range"] = [[1, GVmax]]

            logging.info('Checking GV range for user material properties')
            n_propfile = 0
            for GVrange in matprop["range"]:
                prop_GVmin, prop_GVmax = int(GVrange[0]), int(GVrange[1])
                # checks on property range Min and Max
                if prop_GVmin < GVmin:
                    raise IOError('property GV range below min representable integer')
                    # exit(1)

                if prop_GVmax > GVmax:
                    raise IOError('property GV range exceeds max representable integer')
                    # exit(1)

                prop_GV[n_propfile] = (prop_GVmin, prop_GVmax)
                prop_GV2[n_propfile] = np.arange(prop_GVmin, prop_GVmax + 1, 1)
                n_propfile = n_propfile + 1

            # check prop_GV overlap
            for i in range(n_propfile - 1):
                if not set(prop_GV2[i]).isdisjoint(prop_GV2[i + 1]):
                    raise IOError('GV ranges assigned to different material properties cannot overlap')
                    # exit(1)
        else:
            logging.warning('matprop not given: material property mapping disabled.')

    # write ABAQUS *.inp output file ############################################
    logging.info('Start writing INP file')
    # open ABAQUS *.INP output file
    INP = open(fileout, 'w')

    # HEADER
    INP.write('** ---------------------------------------------------------\n')
    INP.write('** Abaqus .INP file written by Voxel_FEM script on {}\n'.format(datetime.now()))
    INP.write('** ---------------------------------------------------------\n')

    # NODE COORDINATES
    logging.info('Writing model nodes to INP file')
    INP.write('** Node coordinates from input model\n')
    INP.write('*NODE\n')

    # write only the existing nodes
    existing_nodes = np.unique(mesh.cells_dict["hexahedron"])
    for node in existing_nodes:
        INP.write('{0:10d}, {n[0]:12.6f}, {n[1]:12.6f}, {n[2]:12.6f}\n'.format(node, n=mesh.points[node]))

    # ELEMENTS AND ELEMENT SETS
    n_els = 0
    logging.info('Writing model elements to INP file')
    INP.write('** Elements and Element sets from input model\n')
    i = 0

    # for each existing model Grey Value
    for GV in existing_GV:
        # compose element SET definition string and write to output file
        elset_str = 'SET' + str(int(GV))
        INP.write('*ELEMENT, TYPE={0}, ELSET={1}\n'.format(eltype, elset_str))

        # for each cell with that Grey Value
        for cell in np.where(cell_GV == GV)[0]:
            # write cell index followed by a list of its nodes
            INP.write('{0},{n[0]},{n[1]},{n[2]},{n[3]},{n[4]},{n[5]},{n[6]},{n[7]}\n'.format(cell+1, n=mesh.cells[0][1][cell]))

    # NODE SETS
    if 'NSET' in keywords:
        INP.write('** Additional nset from voxel model. New generated nsets are:\n')
        INP.write('** {}, \n'.format(*list(mesh.point_sets)))

        # write node set string
        for nsetName in list(mesh.point_sets):
            INP.write('*NSET, NSET={0}\n'.format(nsetName))
            CR = 1 # carriage return
            # write node set indexes in lines of 10
            for node_i in mesh.point_sets[nsetName]:
                if CR < 10:
                    INP.write('{},'.format(node_i))
                else:
                    INP.write('{}\n'.format(node_i))
                    CR = 0
                CR = CR + 1
            INP.write('\n')
        logging.info('Additional nodes sets generated: {}'.format(list(mesh.point_sets)))

    # CELL SETS
    if 'ELSET' in keywords:
        INP.write('** Additional elset from voxel model. New generated elsets are:\n')
        INP.write('** {}, \n'.format(*list(mesh.cell_sets)))

        # write cell set string
        for csetName in list(mesh.cell_sets):
            INP.write('*ELSET, ELSET={}\n'.format(csetName))
            CR = 1 # carriage return

            # write cells set indexes in lines of 10
            for cell_i in mesh.cell_sets[csetName]:
                if CR < 10:
                    INP.write('{},'.format(cell_i))
                else:
                    INP.write('{}\n'.format(cell_i))
                    CR = 0
                CR = CR + 1
            INP.write('\n')
        logging.info('Additional cell sets generated: {}'.format(list(mesh.cell_sets)))

    # MATERIAL MAPPING
    if 'PROPERTY' in keywords:
        logging.info('User material properties defined. Writing material property section of INP file')
        INP.write('** User material property definition:\n')
        INP.write('** internal variables are: "SetName", "MatName", "GV"\n')

        n_propfile = 0
        for input_property in matprop["file"]:
            # open material property template file
            try:
                PROPfile = open(input_property, 'r')
            except IOError('Material property file not found'):
                exit(1)

            lines = PROPfile.readlines()
            # for each GV corresponding to an existing material index
            for GV in existing_GV:

                # if the GV belongs to the GVrange assigned to the specific material property
                if prop_GV[n_propfile][0] <= GV <= prop_GV[n_propfile][1]:
                    matset_str = 'SET' + str(GV)

                    for line in lines:
                        line = line.replace('\n', '')

                        # copy comment lines as they are
                        if line.startswith('**'):
                            INP.write('{}\n'.format(line))
                            continue

                        # replace keyword 'SetName' and assign current material to the corresponding (by GV) elset
                        if 'SetName' in line:
                            line = line.replace('SetName', matset_str)

                        # replace keyword 'MatName' and create new material flag for current GV
                        if 'MatName' in line:
                            line = line.replace('MatName', 'MAT' + matset_str)

                        # replace keyword 'GV' with current GV and evaluate expression for each material property
                        if 'GV' in line:
                            prop_strings = line.split(',')
                            prop_n = 0
                            for prop in prop_strings:
                                if 'GV' in prop:
                                    INP.write('{}'.format(eval(prop)))
                                else:
                                    INP.write('{}'.format(prop))
                                if prop_n < len(prop_strings) - 1:
                                    INP.write(',')
                                prop_n = prop_n + 1

                            INP.write('\n')
                        else:
                            # copy line as is
                            INP.write('{}\n'.format(line))
            PROPfile.close()
            n_propfile = n_propfile + 1

    # BOUNDARY CONDITIONS AND ANALYSIS DEFINITION
    # Open Abaqus analysis template file
    try:
        template = open(templatefile, 'r')
    except IOError('Abaqus template file {} not found.'.format(templatefile)):
        exit(1)
    logging.info('Reading Abaqus template file {}'.format(templatefile))

    # copy line by line info on model solution and boundary conditions from Abaqus analysis template file
    for line in template.readlines():

        if refnode is not None:
            # replace keywords 'refnodeX', 'refnodeY' and 'refnodeZ' with refnode coordinates
            line = line.replace('refnodeX', str(round(refnode[0], 4)))
            line = line.replace('refnodeY', str(round(refnode[1], 4)))
            line = line.replace('refnodeZ', str(round(refnode[2], 4)))

        # write line to output Abaqus file
        INP.write('{}'.format(line))

    template.close()
    INP.close()
    logging.info('Model with {0} nodes and {1} elements written to file {fname}'.format(len(mesh.points), len(mesh.cells[0]), fname=fileout))

    return

def matpropdictionary(proplist):
    """Compose dictionary of material properties and property mapping GV ranges.

    Parameters
    ----------
    proplist
        List of material property files followed by the corresponding Gray Value range for material mapping.

    Returns
    -------
    matprop : dict
        Dictionary of material properties for material property mapping:
        matprop = {
            "file": ["prop.inp", "property_temp_bone.inp", ...],
            "range": [[250, 255], [0, 250], ...],
        }
    """

    if proplist is None:
        return None

    if len(proplist) == 0:
        return None

    elif len(proplist) == 1:
        if not os.path.isfile(proplist[0]):
            raise IOError('Property file {0} doesn''t exist.', format(proplist[0]))
        else:
            matprop = {"file": [proplist[0]]}
            return matprop

    else:
        # check correct usage of --mapping flag
        if not len(proplist) % 3 == 0:
            raise IOError('Mapping dictionary error (probably due to incorrect use of --mapping flag): each material property file must be followed by MIN and MAX of the corresponding GV range.')

        else:
            matprop = {
                'file': [],
                'range': []
            }
            for prop in range(0, int(len(proplist)/3)):
                matprop['file'].append(proplist[3*prop])
                matprop['range'].append([int(proplist[3*prop+1]), int(proplist[3*prop+2])])

            return matprop



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
MicroCT image processing utilities.

"""

__author__ = 'Gianluca Iori'
__date_created__ = '2021-03-28'
__date__ = '2023-01-14'
__copyright__ = 'Copyright (c) 2022, ORMIR'
__docformat__ = 'restructuredtext en'
__license__ = "MIT"
__version__ = "1.4"
__maintainer__ = 'Gianluca Iori'
__email__ = "gianthk.iori@gmail.com"

import numpy as np
import png
import os
import tifffile
import matplotlib.pyplot as plt

def touint8(data_3D, range=None, quantiles=None, numexpr=True):
    """Normalize and convert data to uint8.

    Parameters
    ----------
    data_3D
        Input data.
    range : [float, float]
        Control range for data normalization.
    quantiles : [float, float]
        Define range for data normalization through input data quantiles. If range is given this input is ignored.
    numexpr : bool
        Use fast numerical expression evaluator for NumPy (memory expensive).

    Returns
    -------
    output : uint8
        Normalized data.
    """

    def convert8bit():

        data_3D_float = data_3D.astype(np.float32, copy=False)
        df = np.float32(data_max - data_min)
        mn = np.float32(data_min)

        if numexpr:
            import numexpr as ne

            scl = ne.evaluate('0.5+255*(data_3D_float-mn)/df', truediv=True)
            ne.evaluate('where(scl<0,0,scl)', out=scl)
            ne.evaluate('where(scl>255,255,scl)', out=scl)
            return scl.astype(np.uint8)
        else:
            data_3D_float = 0.5 + 255 * (data_3D_float - mn) / df
            data_3D_float[data_3D_float < 0] = 0
            data_3D_float[data_3D_float > 255] = 255
            return np.uint8(data_3D_float)

    if range == None:

        # if quantiles is empty data is scaled based on its min and max values
        if quantiles == None:
            data_min = np.nanmin(data_3D)
            data_max = np.nanmax(data_3D)
            data_max = data_max - data_min
            return convert8bit()
        else:
            [data_min, data_max] = np.quantile(np.ravel(data_3D), quantiles)
            return convert8bit()

    else:
        # ignore quantiles input if given
        if quantiles is not None:
            print('quantiles input ignored.')

        data_min = range[0]
        data_max = range[1]
        return convert8bit()

def to01(data_3D):
    """Normalize data to 0-1 range.

    Parameters
    ----------
    data_3D
        Input data.

    Returns
    -------
    data_3D : float32
        Normalized data.
    """
    import numexpr as ne

    data_3D = data_3D.astype(np.float32, copy=False)
    data_min = np.nanmin(data_3D)
    data_max = np.nanmax(data_3D)
    df = np.float32(data_max - data_min)
    mn = np.float32(data_min)
    scl = ne.evaluate('(data_3D-mn)/df', truediv=True)
    return scl.astype(np.float32)

def writemidplanes(data_3D, fileout, slice_x=-1, slice_y=-1, slice_z=-1):
    """Plot orthogonal mid-planes through 3D dataset and save them as images.
    Uses pypng for writing .PNG files.

    Parameters
    ----------
    data
        Input 3D image data.
    fileout : str
        Output .PNG image file name.
    slice_x : int
        X-slice number.
    slice_y : int
        Y-slice number.
    slice_z : int
        Z-slice number.
    """

    if data_3D.ndim == 3:

        if slice_x == -1:
            slice_x = int(data_3D.shape[2] / 2)
        if slice_y == -1:
            slice_y = int(data_3D.shape[1] / 2)
        if slice_z == -1:
            slice_z = int(data_3D.shape[0] / 2)

        filename, ext = os.path.splitext(fileout)
        with open(filename + '_XY.png', 'wb') as midplaneXY:
            pngWriter = png.Writer(data_3D.shape[2], data_3D.shape[1], greyscale=True, alpha=False, bitdepth=8)
            pngWriter.write(midplaneXY, touint8(data_3D[int(slice_z), :, :]))

        with open(filename + '_XZ.png', 'wb') as midplaneXZ:
            pngWriter = png.Writer(data_3D.shape[2], data_3D.shape[0], greyscale=True, alpha=False, bitdepth=8)
            pngWriter.write(midplaneXZ, touint8(data_3D[:, int(slice_y), :]))

        with open(filename + '_YZ.png', 'wb') as midplaneYZ:
            pngWriter = png.Writer(data_3D.shape[1], data_3D.shape[0], greyscale=True, alpha=False, bitdepth=8)
            pngWriter.write(midplaneYZ, touint8(data_3D[:, :, int(slice_x)]))

def plot_midplanes(data_3D, slice_x=-1, slice_y=-1, slice_z=-1):
    """Plot orthogonal cross-sections through 3D dataset.

    Parameters
    ----------
    data_3D
        Input 3D image data.
    slice_x : int
        X-slice number.
    slice_y : int
        Y-slice number.
    slice_z : int
        Z-slice number.
    """

    if slice_x == -1:
        slice_x = int(data_3D.shape[2] / 2)
    if slice_y == -1:
        slice_y = int(data_3D.shape[1] / 2)
    if slice_z == -1:
        slice_z = int(data_3D.shape[0] / 2)

    fig, (ax1, ax2, ax3) = plt.subplots(1, 3)
    ax1.imshow(data_3D[slice_z, :, :])
    ax2.imshow(data_3D[:, slice_y, :])
    ax3.imshow(data_3D[:, :, slice_x])

def plot_projections(data_3D, projection='max'):
    """Plot orthogonal projections of 3D dataset.

    Parameters
    ----------
    data_3D
        Input 3D image data.
    projection : str
        Projection method. Available choices are 'max', 'min'.
    """

    fig, (ax1, ax2, ax3) = plt.subplots(1, 3)

    if projection=='max':
        ax1.imshow(np.max(data_3D, 0))
        ax2.imshow(np.max(data_3D, 1))
        ax3.imshow(np.max(data_3D, 2))
    elif projection=='min':
        ax1.imshow(np.min(data_3D, 0))
        ax2.imshow(np.min(data_3D, 1))
        ax3.imshow(np.min(data_3D, 2))

def read_tiff_stack(filename, range=None, zfill=4):
    """Read stack of tiff files. Searches all files in parent folder and opens them as a stack of images.

    Parameters
    ----------
    filename
        One of the stack images.
    range : [int, int]
        Control load slices range.
    zfill : int
        Number of leading zeros in file names.

    TO DO:
    ----------
    - check that folder contains only .TIFF files; skip the rest
    """

    # search all files in parent folder; create filenames list
    stack_files = [os.path.join(os.path.dirname(filename), f) for f in os.listdir(os.path.dirname(filename))
                     if os.path.isfile(os.path.join(os.path.dirname(filename), f))]
    stack_files.sort()

    if range is not None:
        import re
        slice_in = [i for i, item in enumerate(stack_files) if re.search(str(range[0]).zfill(4) + ".", item)]
        slice_end = [i for i, item in enumerate(stack_files) if re.search(str(range[1]).zfill(4) + ".", item)]

        if len(slice_in) == 1 and len(slice_end) == 1:
            stack_files = stack_files[slice_in[0]:slice_end[0]]
        else:
            import warnings
            warnings.warn('Given slice range is ambiguous or non existing.. loading whole stack.')

    # load stack using tifffile
    return tifffile.imread(stack_files)

def bbox(bw, pad=0, dsize=None, verbose=None):
    """Bounding BOX limits of input binary image.

    Parameters
    ----------
    bw : bool
        Binary image.
    pad : int
        Add padding of given number of pixels to the BBOX limits.
    dsize : int
        perform image close with disk structuring element of radius 'dsize' before calculating the BBOX.
    verbose
        Activate verbose graphical output

    Returns
    -------
    bbox_origin: int
        Origin [row col (slice)] of the BBOX inscribing True values in input image bw.
    bbox_size: int
        BBOX size [s_row s_col (s_slice)].
    """

    # DSIZE: remove artefacts > erode/dilate
    if dsize:
        raise IOError('dsize method not implemented yet.')

    if bw.ndim == 3:
        # project along each dimension
        maxROW = np.max(np.max(bw, 0), 1)
        maxCOL = np.max(np.max(bw, 0), 0)
        maxSLICE = np.max(np.max(bw, 1), 1)

        # find first and last True occurrences
        row0 = list(maxROW).index(True)
        row1 = len(maxROW) - list(maxROW[::-1]).index(True) - 1

        col0 = list(maxCOL).index(True)
        col1 = len(maxCOL) - list(maxCOL[::-1]).index(True) - 1

        slice0 = list(maxSLICE).index(True)
        slice1 = len(maxSLICE) - list(maxSLICE[::-1]).index(True) - 1

        # add padding
        row0 = row0 - pad
        rowd = row1 - row0 + pad
        col0 = col0 - pad
        cold = col1 - col0 + pad
        slice0 = slice0 - pad
        sliced = slice1 - slice0 + pad

        if pad > 0:
            # check if bbox exceeds image size
            if row0 < 0:
                row0=0
            if col0 < 0:
                col0=0
            if slice0 < 0:
                slice0=0

            bw_size = bw.shape
            if slice0 + sliced > bw_size[0]:
                sliced = bw_size[0] - slice0
            if row0 + rowd > bw_size[1]:
                rowd = bw_size[1] - row0
            if col0 + cold > bw_size[2]:
                cold = bw_size[2]-col0

        if verbose:
            fig, (ax1, ax2) = plt.subplots(1, 2)
            ax1.imshow(np.max(bw, 0))
            ax1.plot([col0, col0], [0, bw.shape[1]-1], 'r')
            ax1.plot([col0+cold, col0+cold], [0, bw.shape[1]-1], 'r')
            ax1.plot([0, bw.shape[2] - 1], [row0, row0], 'r')
            ax1.plot([0, bw.shape[2] - 1], [row0 + rowd, row0 + rowd], 'r')

            ax2.imshow(np.max(bw, 1))
            ax2.plot([col0, col0], [0, bw.shape[1] - 1], 'r')
            ax2.plot([col0 + cold, col0 + cold], [0, bw.shape[1] - 1], 'r')
            ax2.plot([0, bw.shape[0] - 1], [slice0, slice0], 'r')
            ax2.plot([0, bw.shape[0] - 1], [slice0 + sliced, slice0 + sliced], 'r')

        bbox_origin = [row0, col0, slice0]
        bbox_size = [rowd, cold, sliced]

        return bbox_origin, bbox_size

    if bw.ndim == 2:
        raise IOError('bbox method for 2D images not implemented yet.')

def crop(data_3D, crop_origin, crop_size):
    """Crop 3D image given crop origin and size.

    Parameters
    ----------
    data_3D
        Input data.
    crop_origin : [int, int, int]
        Crop origin [Z,Y,X].
    crop_size : [int, int, int]
        Crop size [Z,Y,X].

    Returns
    -------
    output
        Cropped data.
    """
    return data_3D[crop_origin[2]:crop_origin[2] + crop_size[2], crop_origin[0]:crop_origin[0] + crop_size[0], crop_origin[1]:crop_origin[1] + crop_size[1]]


#!/usr/bin/python
"""
This script extracts data from ccx dat-files.
Extended from the dat2txt.py helper script by
https://github.com/mkraska/CalculiX-Examples/tree/master/Scripts

"""
import re
import glob
import argparse
import logging
import textwrap

def dat2txt(filename, verbose=False):
    """Read CalculiX .DAT output file and write analysis results to .TXT file.

    Parameters
    ----------
    filename
        CalculiX ouput .DAT file.
    verbose
        Verbose output.

    Returns
    -------
    resname
        results name
    group
        results group name
    time_points
        number of time steps
    """

    if verbose:
        logging.basicConfig(level=logging.INFO)

    # variables initialization
    resname = ''
    group = ''
    data = {}
    pH = re.compile(' (.+) for .*set\\s(\\S+) and time  (.+)')
    skip = 0  # if empty lines are to be skipped
    body = 0  # if data lines are expected
    nev = 0  # number of eigenvalue file
    stop_on_empty = 0  # empty line triggers end of body
    time_points = 0

    # open job file
    logging.info('Reading file {}.'.format(filename))
    f = open(filename, "r")

    for line in f:
        if skip:  # if the line is known to be useless
            skip = skip - 1
            continue
        line = line.replace("(", "")
        line = line.replace(")", "")
        line = line.replace("\n", "")
        b = pH.match(line)
        if b:  # a result header was found
            resname = b.group(1)
            group = b.group(2)
            # print(group)
            res = resname + "_" + group
            time = float(b.group(3))
            if not (res in data.keys()):  # new result type
                res_filename = res + ".txt"
                data[res] = open(res_filename, "w")
            data[res].write("\n" + str(time) + " ")
            line_end = " "
            body = 1
            stop_on_empty = 0
        elif line.startswith("     E I G E N V A L U E   O U T P U T"):
            # eigenvalue data
            nev += 1
            res = "Eigenvalues_" + str(nev)
            # print(res)
            data[res] = open(res + ".txt", "w")
            data[res].write("# mode  EV  Re(f)_rad/time Re(f)_cycles/time Im(f)_rad/time\n")
            line_end = "\n"
            body = 1
            skip = 5
            stop_on_empty = 1
        elif line.startswith("     P A R T I C I P A T I O N   F A C T O R S"):
            # eigenvalue data
            res = "Eigenvalues_PF_" + str(nev)
            print(res)
            data[res] = open(res + ".txt", "w")
            data[res].write("# mode  UX  UY UZ RX RY RZ\n")
            line_end = "\n"
            body = 1
            skip = 3
            stop_on_empty = 1
        elif line.startswith("     E F F E C T I V E   M O D A L   M A S S"):
            # eigenvalue data
            res = "Eigenvalues_MM_" + str(nev)
            print(res)
            data[res] = open(res + ".txt", "w")
            data[res].write("# mode  UX  UY UZ RX RY RZ\n")
            line_end = "\n"
            body = 1
            skip = 3
            stop_on_empty = 1
        else:
            if body == 1:
                if not "force" in line and not "center" in line:
                    data[res].write(line + line_end)
                    if line.strip():
                        time_points += 1
                if stop_on_empty and line == "":
                    body = 0

    for name in data.keys():
        data[name].close()

    logging.info('Written {} for group {} to file: {}.\nTotal time points: {}.\n'.format(resname, group, res_filename, time_points))
    return resname, group, time_points

def main():
    description = textwrap.dedent('''\
                Script to process CalculiX Finite Element (FE) output .DAT files.
                Reads CalculiX FE analysis results and writes a .TXT file of the output variables and time steps.
                ''')

    parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('job', type=str, help='Input CalculiX job name.')
    # parser.add_argument('j', '--jobname', type=str, default=None, help='<CalculiX job name.')
    # parser.add_argument('fileout', type=str, default=None, help='<Required> Output filename (Abaqus .INP).')
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='Verbose output.')
    parser.set_defaults(job=None, verbose=True)

    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.INFO)

    # processing command line arguments, get the jobname
    if args.job is None:
        logging.info("No jobname given.\n")
        files = glob.glob("*.dat")
        if len(files) == 1:
            logging.info('File {} found'.format(files[0]))
            job = files[0]
        else:
            logging.exception('Available data files: {}'.format(*f) for f in files)
            quit()
    else:
        job = args.job + ".dat"

    # run dat2txt
    resname, group, time_points = dat2txt(job)

if __name__ == '__main__':
    main()

# names = ['time', 'fx', 'fy', 'fz']
# filename = '/home/gianthk/PycharmProjects/CT2FE/test_data/steel_foam/total force fx,fy,fz_NODES_B.txt'
# data2 = pd.read_table(filename, delim_whitespace=True, names=names, index_col=0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
From 3D Computed Tomography (CT) images to Finite Element (FE) models.
For more information, call this script with the help option:
    ciclope.py -h

"""

__author__ = ['Gianluca Iori', 'Martino Pani']
__date_created__ = '2021-08-06'
__date__ = '2022-08-14'
__copyright__ = 'Copyright (c) 2022, ORMIR'
__docformat__ = 'restructuredtext en'
__license__ = "GPL"
__version__ = "1.1.2"
__maintainer__ = 'Gianluca Iori'
__email__ = "gianthk.iori@gmail.com"

import os
import argparse
import logging
import textwrap
import numpy as np
import meshio
from skimage.filters import gaussian
from ciclope.core import voxelFE, tetraFE
from ciclope.utils import preprocess, recon_utils

#################################################################################

def main():
    description = textwrap.dedent('''\
                Ciclope generates CalculiX Finite Element (FE) input files from 3D voxel data.

                The output of this script is an input file (.INP) in ABAQUS syntax
                that can be solved using ABAQUS or CALCULIX.

                The input 3D image is segmented and processed to maintain only the largest connected structure.
                3D voxel or tetrahedra FE input files can be generated.
                Meshes for 3D rendering of the outer model shell and of the whole selected volume can be generated.

                The user can define a material mapping strategy (implemented only for voxel FE)
                for the direct conversion of local Grey Values (GVs) of the 3D image
                to local material properties of the FE model.

                The definition of material mapping laws is done using separate template file(s).
                See the folder "material_properties" for examples.

                Boundary conditions (BCs), analysis steps and associated output requests
                are defined in a separate template file. See the folder "input_templates" for examples.
                More info on how to define analysis steps in ABAQUS at:
                https://abaqus-docs.mit.edu/2017/English/SIMACAECAERefMap/simacae-m-Sim-sb.htm#simacae-m-Sim-sb
                ''')
    epilog = textwrap.dedent('''\
                EXAMPLES:

                * Convert SRCT scan of single strut to voxel FE model.
                  GVs between 1 and 249 are used for mapping local bone tissue material properties.
                  GVs between 250 and 255 are reserved for a steel cap, and a separate property file is given
                  for the mechanical properties of steel. A static tensile analysis and output requests are
                  defined in the ABAQUS template file "tmp_example02_tens_static.inp":

                    ciclope.py /home/gianthk/Data/TOMCAT/Kaya/D_single_h1h2_scale05/D_single_h1h2_scale050017.tif pippo.inp
                    -vs 3.25 3.25 3.25
                    -r 5
                    --voxelfe
                    --template /home/gianthk/PycharmProjects/CT2FE/input_templates/tmp_example01_tens_static.inp
                    --mapping
                    /home/gianthk/PycharmProjects/CT2FE/material_properties/steel.inp 250 255
                    /home/gianthk/PycharmProjects/CT2FE/material_properties/bone.inp 1 249

                * Same as the example above, but without material mapping.
                  The image is segmented and all voxels belonging to the largest connected strut are assigned
                  to the same elements set with uniform material properties. The analysis template file 
                   is assumed to contain the material property definition (see file "tmp_example02_tens_static.inp"):

                    ciclope.py /home/gianthk/Data/TOMCAT/Kaya/D_single_h1h2_scale05/D_single_h1h2_scale050017.tif pippo.inp
                    -vs 3.25 3.25 3.25
                    -r 5
                    --voxelfe
                    --template /home/gianthk/PycharmProjects/CT2FE/input_templates/tmp_example02_tens_static.inp
                ''')
    onemoreexample = textwrap.dedent('''
                * Convert CT scan of embedded bone tissue mapping the model GVs to local material properties.
                  GVs between 251 and 255 are reserved for the embedding matrix, and a separate property file is given
                  for the mechanical properties of this material. The type of FE analysis, all BCs and requested output are
                  defined in the template file "tmp.inp":

                    ciclope.py scan-sample_15-3-crop_c0001.tif pippo.inp
                    -r 4
                    --vol_mesh
                    --shell_mesh
                    --voxelfe
                    --template /home/gianthk/PycharmProjects/CT2FE/input_templates/tmp_example02_tens_static.inp
                    ''')

    parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('filein', type=str, help='<Required> Input filename (voxel data).')
    parser.add_argument('fileout', type=str, default=None, help='<Required> Output filename (Abaqus .INP).')
    parser.add_argument('-vs', '--voxelsize', type=float, default=[1., 1., 1.], nargs='+', help='Voxel size.')
    parser.add_argument('-r', '--resampling', type=float, default=1., help='Resampling factor.')
    parser.add_argument('-t', '--threshold', type=int, default=None, help='Threshold value.')
    parser.add_argument('-s', '--smooth', type=float, nargs='?', const=1., default=0.,
                        help='Smooth image with gaussian filter of given Sigma before thresholding.')
    parser.add_argument('--caps', type=int, default=None,
                        help='Add caps of given thickness to the bottom and top of the model for mesh creation.')
    parser.add_argument('--caps_val', type=int, default=0, help='Caps grey value.')
    parser.add_argument('--shell_mesh', dest='shell_mesh', action='store_true',
                        help='Write VTK mesh of outer shell generated with PyMCubes.')
    parser.add_argument('--vol_mesh', dest='vol_mesh', action='store_true',
                        help='Write VTK volume mesh of tetrahedra with pygalmesh.')
    parser.add_argument('--max_facet_distance', type=float, default=None, help='CGAL mesh parameter.')
    parser.add_argument('--max_cell_circumradius', type=float, default=None, help='CGAL mesh parameter.')
    parser.add_argument('--voxelfe', dest='voxelfe', action='store_true', help='Write voxel FE model (.INP) file.')
    parser.add_argument('--template', type=str, default=None,
                        help='<Required by --voxelfe> Abaqus analysis template file (.INP).')
    parser.add_argument('-m', '--mapping', default=None, nargs='+',
                        help='Template file for material property mapping. If more than one property is given, each property filename must followed by the corresponding GV range.')
    parser.add_argument('--tetrafe', dest='tetrafe', action='store_true',
                        help='Write linear tetrahedra FE model (.INP) file.')
    parser.add_argument('--refnode', default=None, nargs='+',
                        help='Reference node input. Used for kinematic coupling of Boundary Conditions in the analysis template file.'
                             'The REF_NODE coordinates [x,y,z] can be given. Alternatively use one of the following args [X0, X1, Y0, Y1, Z0, Z1] to generate a REF_NODE at a model boundary.')
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='Verbose output.')
    parser.set_defaults(shell_mesh=False, vol_mesh=False, voxelfe=False, tetrafe=False, verbose=False)

    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.INFO)

    # filename base
    [fileout_base, type] = os.path.splitext(args.fileout)

    # isotropic voxel size (spacing) [mm]
    sp = np.array(args.voxelsize)

    # Read tiff stack #######################################################
    I = recon_utils.read_tiff_stack(args.filein)

    # Gaussian smooth #######################################################
    if args.smooth != 0:
        logging.info('Smoothing image with gaussian kernel')
        I = gaussian(I, sigma=args.smooth, preserve_range=True)

    # Resize the dataset ####################################################
    # use scikit
    if args.resampling != 1:
        logging.info("Resampling input dataset with factor 1/{}".format(args.resampling))
        I, sp = preprocess.resample(I, sp, args.resampling)
        logging.info("New voxelsize: {}".format(sp))

    if args.verbose:
        # plot the midplanes through the stack
        recon_utils.plot_midplanes(I)
        # write midplanes as .PNG
        recon_utils.writemidplanes(I, fileout_base + ".png")

    # Add image caps before thresholding ####################################
    if args.caps is not None:
        # add cap so that the mesh is closed
        I = recon_utils.add_cap(I, cap_thickness=args.caps, cap_val=args.caps_val)

    # Binarise the dataset ##################################################
    BW, T = preprocess.segment(I, args.threshold)
    if args.threshold is None:
        logging.info("Threshold value not given. Using Otsu method..")

    logging.info("Threshold: {}".format(T))

    # Keep largest isolated cluster of voxels #################################
    logging.info("Removing unconnected clusters of voxels..")
    L = preprocess.remove_unconnected(BW)

    # Visualization with Napari #############################################
    # import napari
    # viewer = napari.view_image(L)

    # Generate outer shell mesh #############################################
    if args.shell_mesh:
        logging.info("Writing triangle mesh of outer shell")
        # shell mesh using pymcubes (high resolution mesh; caps excluded)
        vertices, triangles, shellmesh = tetraFE.shell_mesh(L, method='pymcubes')

        # write VTK mesh with meshio
        meshio.write_points_cells(fileout_base + "_shell.vtk", vertices.tolist(), [("triangle", triangles.tolist())])

        # shell mesh using pygalmesh (CGAL) (low resolution mesh; caps included)
        # check CGAL parameters
        # max_facet_distance, max_cell_circumradius = check_cgal_params(args.max_facet_distance, args.max_cell_circumradius, sp)

        # vertices, triangles, shellmesh = shell_mesh(L, 'pygalmesh', sp, max_facet_distance, max_cell_circumradius)
        # shellmesh.write(fileout_base + "_shell.vtk")

    # Generate volume mesh ##################################################
    if args.vol_mesh:
        # check CGAL parameters
        max_facet_distance, max_cell_circumradius = tetraFE.check_cgal_params(args.max_facet_distance,
                                                                              args.max_cell_circumradius, sp)

        # generate mesh with CGAL
        volmesh = tetraFE.cgal_mesh(L, sp, 'tetra', max_facet_distance, max_cell_circumradius)

        # write the mesh to file
        logging.info("Writing tetrahedra volume mesh")
        volmesh.write(fileout_base + "_tetramesh.vtk")

    # Generate voxel FE model ##################################################
    if args.voxelfe:
        # check voxelfe args
        if not args.template:
            raise IOError('Analysis template file (--template) required with --voxelfe.')

        # build dictionary of material properties
        matprop = voxelFE.matpropdictionary(args.mapping)

        # generate unstructured grid mesh from 3D data
        mesh, refnodes = voxelFE.vol2ugrid(L, sp, refnodes=True)

        if args.refnode is not None:
            if args.refnode[0] in refnodes:
                refnode = refnodes[args.refnode[0]]
            elif len(args.refnode) > 1:
                refnode = args.refnode
            else:
                logging.warning('Refnode not recognized. Option deactivated.')
                refnode = None
        else:
            refnode = args.refnode

        if args.mapping is None:
            # voxelFE of binary volume data; the material property definition is assumed to be in the analysis template file
            voxelFE.mesh2voxelfe(mesh, args.template, fileout_base + "_voxelFE.inp", keywords=['NSET', 'ELSET'], refnode=refnode, verbose=args.verbose)

        else:
            # mask the data (keep only the largest connected volume)
            masked = I
            masked[~L] = 0
            # voxelFE of greyscale volume data; material mapping on
            voxelFE.mesh2voxelfe(mesh, args.template, fileout_base + "_voxelFE.inp", matprop, keywords=['NSET', 'ELSET', 'PROPERTY'], refnode=refnode, verbose=args.verbose)

    # Generate tetrahedra FE model #############################################
    if args.tetrafe:
        # check tetrafe args
        if not args.template:
            raise IOError('Analysis template file (--template) required with --tetrafe.')

        # # build dictionary of material properties
        # matprop = matpropdictionary(args.mapping)

        if args.mapping is None:
            if 'volmesh' not in locals():
                # check CGAL parameters
                max_facet_distance, max_cell_circumradius = tetraFE.check_cgal_params(args.max_facet_distance, args.max_cell_circumradius, sp)

                # generate volume mesh
                volmesh = tetraFE.cgal_mesh(L, sp, 'tetra', max_facet_distance, max_cell_circumradius)

            elif volmesh is None:
                # check CGAL parameters
                max_facet_distance, max_cell_circumradius = tetraFE.check_cgal_params(args.max_facet_distance, args.max_cell_circumradius, sp)

                # generate volume mesh
                volmesh = tetraFE.cgal_mesh(L, sp, 'tetra', max_facet_distance, max_cell_circumradius)

            # tetraFE of the already meshed volume; the material property definition is assumed to be in the analysis template file
            tetraFE.mesh2tetrafe(volmesh, args.template, fileout_base + "_tetraFE.inp", keywords=['NSET', 'ELSET'],
                                 verbose=args.verbose)

        else:
            logging.exception("--tetrafe with mapping option not implemented yet.")

    return

if __name__ == '__main__':
    main()





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Ciclope postprocessing module
"""

import sys
sys.path.append('~/Applications/ParaView-5.9.0-RC1-MPI-Linux-Python3.8-64bit/lib/python3.8/site-packages')

try:
    from paraview.simple import *
except ImportError:
    import warnings
    warnings.warn("ParaView.simple is required!", RuntimeWarning)

def paraview_plot(filein, fileout=None, slicenormal="XYZ", RepresentationType="Surface", Crinkle=False, ColorBy="S_Mises", Roll=0, ImageResolution=[1280, 960], TransparentBackground=True, ColorMap='Viridis (matplotlib)'):
    """Plot field data using ParaView.
    ParaView must be installed and a link to its python library must be added to your system path.

    Parameters
    ----------
    filein : str
        Input data (VTK or other 3D ParaView file).
    fileout : str
        Output image file name.
    slicenormal : str
        Any combination of 'X', 'Y', and 'Z'. Default='XYZ'.
    RepresentationType : str
        'Surface', 'SurfaceWithEdges', 'Volume', 'Points', 'Feature Edges', or '3D Glyphs'. Default='Surface'.
    Crinkle : bool
        Crinkle the slice. Default=False.
    ColorBy : str
        Field name for coloring. Default='S_Mises'
    Roll : int
        View roll angle. Default=0.
    ImageResolution : int
        Output image resolution [X, Y] in pixels. Default=[1280, 960].
    TransparentBackground : bool
        Transparent background. Default=True.
    Colormap : str
        Default = Viridis.
    """

    # read vtk file
    data = LegacyVTKReader(FileNames=filein)

    slice = Slice(registrationName='Slice1', Input=data)
    slice.SliceType = 'Plane'
    slice.HyperTreeGridSlicer = 'Plane'

    # get active view
    renderView1 = GetActiveViewOrCreate('RenderView')

    # draw the object
    # show data in view
    slice1Display = Show(slice, renderView1, 'GeometryRepresentation')

    # get the object bounds
    bounds = data.GetDataInformation().GetBounds()
    center = [(bounds[1] - bounds[0]) / 2, (bounds[3] - bounds[2]) / 2, (bounds[5] - bounds[4]) / 2]

    slice.SliceOffsetValues = [0.0]

    # slice.SliceType.Origin = [2., 2., 1.4]
    slice.SliceType.Origin = center

    colorby_string = ''.join(ColorBy)

    if fileout is None:
        import os
        filename_out_base, ext_out = os.path.splitext(filein)

    if 'x' in slicenormal.lower():
        slice.SliceType.Normal = [1.0, 0.0, 0.0]

        if Crinkle:
            slice.Crinkleslice = 1

        # reset view to fit data
        renderView1.ResetCamera()

        # changing interaction mode based on data extents
        renderView1.InteractionMode = '2D'
        renderView1.CameraPosition = [10000, center[1], center[2]]
        renderView1.CameraFocalPoint = center

        plot_slice(renderView1, slice1Display, filename_out_base +'_' + colorby_string +'_YZ.png', RepresentationType, ColorBy, Roll, ImageResolution, TransparentBackground, ColorMap)

    if 'y' in slicenormal.lower():
        slice.SliceType.Normal = [0.0, 1.0, 0.0]

        if Crinkle:
            slice.Crinkleslice = 1

        # reset view to fit data
        renderView1.ResetCamera()

        # changing interaction mode based on data extents
        renderView1.InteractionMode = '2D'
        renderView1.CameraPosition = [center[0], 10000, center[2]]
        renderView1.CameraFocalPoint = center

        plot_slice(renderView1, slice1Display, filename_out_base + '_' + colorby_string + '_XZ.png', RepresentationType, ColorBy, (-90 + Roll), ImageResolution, TransparentBackground, ColorMap)

    if 'z' in slicenormal.lower():
        slice.SliceType.Normal = [0.0, 0.0, 1.0]

        if Crinkle:
            slice.Crinkleslice = 1

        # reset view to fit data
        renderView1.ResetCamera()

        # changing interaction mode based on data extents
        renderView1.InteractionMode = '2D'
        renderView1.CameraPosition = [center[0], center[1], 10000]
        renderView1.CameraFocalPoint = center

        plot_slice(renderView1, slice1Display, filename_out_base + '_' + colorby_string + '_XY.png', RepresentationType, ColorBy, Roll, ImageResolution, TransparentBackground, ColorMap)

    if not ('x' in slicenormal.lower()) | ('y' in slicenormal.lower()) | ('z' in slicenormal.lower()):
        raise IOError('Invalid Slice Normal.')

    return

def plot_slice(renderView1, slice1Display, fileout, RepresentationType, colorby, Roll, ImageResolution, TransparentBackground, ColorMap):
    """Save plots using Paraview.

    Parameters
    ----------
    renderView1
    slice1Display
    fileout
    RepresentationType
    colorby
    Roll
    ImageResolution
    TransparentBackground
    ColorMap
    """

    import time

    # change representation type
    # slice1Display.SetRepresentationType('Surface')
    slice1Display.SetRepresentationType(RepresentationType)

    # set scalar coloring
    if len(colorby) == 2:
        ColorBy(slice1Display, ('POINTS', colorby[0], colorby[1]))
    else:
        ColorBy(slice1Display, ('POINTS', colorby, colorby))

    # rescale color and/or opacity maps used to include current data range
    slice1Display.RescaleTransferFunctionToDataRange(True, False)

    # show color bar/color legend
    slice1Display.SetScalarBarVisibility(renderView1, True)

    # get color transfer function/color map
    if len(colorby) == 2:
        LUT = GetColorTransferFunction(colorby[0])
    else:
        LUT = GetColorTransferFunction(colorby)

    # get opacity transfer function/opacity map
    if len(colorby) == 2:
        PWF = GetOpacityTransferFunction(colorby[0])
    else:
        PWF = GetOpacityTransferFunction(colorby)

    # Apply a preset using its name. Note this may not work as expected when presets have duplicate names.
    LUT.ApplyPreset(ColorMap, True)

    # get color legend/bar for LUT in view renderView1
    LUTColorBar = GetScalarBar(LUT, renderView1)

    # change scalar bar placement
    # LUTColorBar.Orientation = 'Horizontal'
    LUTColorBar.Orientation = 'Vertical'
    LUTColorBar.WindowLocation = 'UpperRightCorner'  # 'AnyLocation', 'LowerRightCorner', 'LowerLeftCorner', 'LowerCenter', 'UpperLeftCorner', 'UpperRightCorner', 'UpperCenter'
    LUTColorBar.TitleColor = [0, 0, 0]  # switch to black
    LUTColorBar.LabelColor = [0, 0, 0]  # switch to black
    LUTColorBar.TitleFontSize = 10
    LUTColorBar.LabelFontSize = 10
    LUTColorBar.ScalarBarThickness = 8

    # Data Axis visibility
    slice1Display.DataAxesGrid.GridAxesVisibility = 1

    # Data Axis Font
    slice1Display.DataAxesGrid.XTitleFontSize = 40
    slice1Display.DataAxesGrid.XTitleColor = [0.0, 0.0, 0.0]  # switch to black
    slice1Display.DataAxesGrid.XLabelFontSize = 40
    slice1Display.DataAxesGrid.XLabelColor = [0.0, 0.0, 0.0]
    slice1Display.DataAxesGrid.YTitleFontSize = 40
    slice1Display.DataAxesGrid.YTitleColor = [0.0, 0.0, 0.0]  # switch to black
    slice1Display.DataAxesGrid.YLabelFontSize = 40
    slice1Display.DataAxesGrid.YLabelColor = [0.0, 0.0, 0.0]
    slice1Display.DataAxesGrid.ZTitleFontSize = 40
    slice1Display.DataAxesGrid.ZTitleColor = [0.0, 0.0, 0.0]  # switch to black
    slice1Display.DataAxesGrid.ZLabelFontSize = 40
    slice1Display.DataAxesGrid.ZLabelColor = [0.0, 0.0, 0.0]

    slice1Display.DataAxesGrid.GridColor = [0.0, 0.0, 0.0]  # grid color to black

    # update the view to ensure updated data information
    renderView1.Update()

    # dp = GetDisplayProperties()
    #
    # #set point color
    # dp.AmbientColor = [1, 0, 0] #red
    #
    # #set surface color
    # dp.DiffuseColor = [0, 1, 0] #blue
    #
    # #set point size
    # dp.PointSize = 2

    # #set representation
    # dp.Representation = "Surface"

    camera = GetActiveCamera()
    # camera.Elevation(45)
    camera.Roll(Roll)
    Render()

    # save screenshot
    if TransparentBackground:
        SaveScreenshot(fileout, ImageResolution=ImageResolution, TransparentBackground=1)
    else:
        SaveScreenshot(fileout, ImageResolution=ImageResolution)

    time.sleep(1)

    return


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Ciclope image preprocessing module
"""

import numpy as np
from scipy import ndimage, misc
from skimage import measure, morphology
from skimage.filters import threshold_otsu, gaussian
import logging
from tqdm import tqdm
from . import recon_utils as ru

def segment(image, threshold_value):
    """Threshold image.

    Parameters
    ----------
    image
        Image data.
    threshold_value (optional)
        Threshold value. If empty an Otsu threshold is calculated.

    Returns
    -------
    BWimage
        Binary image after thresholding.
    T
        Threshold value.
    """

    # we do want bone = 1 and background = 0;
    if threshold_value is None:
        # use Otsu if threshold input not specified
        T = threshold_otsu(image)

    else:
        T = int(threshold_value)

    # apply the threshold
    return image > T, T

def resample(image, voxelsize, resampling_factor):
    """Resize image.

    Parameters
    ----------
    image
        Image data.
    voxelsize
        Voxel size.
    resampling_factor
        Scaling factor.

    Returns
    -------
    image
        Resized image.
    voxelsize
        Voxel size after rescaling.
    """

    # resize the 3D data using spline interpolation of order 2
    image = ndimage.zoom(image, 1 / resampling_factor, output=None, order=2)

    # correct voxelsize
    voxelsize = voxelsize * resampling_factor

    return image, voxelsize

def centerofmass(bwimage):
    """Center Of Mass (COM) of binary image.

    Parameters
    ----------
    bwimage: bool
        Binary image. Can be 2D and 3D.

    Returns
    -------
    cmassx_array
        X-coordinate array of the COM. If input is 3D, an array of the slicewise COMs is returned.
    cmassy_array
        Y-coordinate array of the COM.
    """

    if bwimage.ndim == 3:

        # output arrays initialization
        cmassx_array = np.zeros([bwimage.shape[0]])
        cmassy_array = np.zeros([bwimage.shape[0]])

        for slice in range(0, bwimage.shape[0]):
            y = np.sum(bwimage[slice,:,:], 1)
            cmassy = np.inner(y, np.arange(0, y.size))
            cmassy_array[slice] = cmassy / np.sum(y)

            x = np.sum(bwimage[slice, :, :], 0)
            cmassx = np.inner(x, np.arange(0, x.size))
            cmassx_array[slice] = cmassx / np.sum(x)

    elif bwimage.ndim == 2:
        y = np.sum(bwimage, 1)
        cmassy = np.inner(y, np.arange(0, y.size))
        cmassy_array = cmassy / np.sum(y)

        x = np.sum(bwimage, 0)
        cmassx = np.inner(x, np.arange(0, x.size))
        cmassx_array = cmassx / np.sum(x)

    return cmassx_array, cmassy_array

def remove_unconnected(bwimage):
    """Remove all unconnected voxels. Returns a binary of the largest connected cluster.

    Parameters
    ----------
    bwimage
        Binary image.

    Returns
    -------
    bwcluster
        Binary image of the largest connected cluster of voxels.
    """

    # label the BW image
    # [labels, n_labels] = measure.label(bwimage, None, True)
    [labels, n_labels] = measure.label(bwimage, None, True, 1)

    # count occurrences of each label
    occurrences = np.bincount(labels.reshape(labels.size))

    # find largest unconnected label
    largest_label_id = occurrences[1:].argmax() + 1

    return labels == largest_label_id

def fill_voids(I, fill_val=None, makecopy=False):
    """Fill voids within color image with given value.

    Parameters
    ----------
    I
        Input color image.
    fill_val
        Filling value.
    makecopy : bool
        Make copy of input image.

    Returns
    -------
    I_filled
        Filled image.
    """

    if fill_val is None:
        fill_val = I.max()

    # binarize and label inverse of the input image
    [labels, n_labels] = measure.label(~(I>0), None, True, 1)

    # count occurrences of each label
    occurrences = np.bincount(labels.reshape(labels.size))

    # find and delete largest label (background)
    largest_label_id = occurrences[1:].argmax() + 1
    labels[labels == largest_label_id] = 0

    if makecopy:
        I_filled = I.copy()
        I_filled[labels != 0] = fill_val
        return I_filled
    else:
        I[labels != 0] = fill_val
        return I

def remove_largest(bwimage):
    """Remove largest cluster of voxels in binary image.

    Parameters
    ----------
    bwimage
        Binary image.

    Returns
    -------
    bwcluster
        Binary image in which the largest cluster of voxels is removed.
    """

    # label the BW image
    [labels, n_labels] = measure.label(bwimage, None, True, 1)

    # count occurrences of each label
    occurrences = np.bincount(labels.reshape(labels.size))

    # find largest unconnected label
    largest_label_id = occurrences[1:].argmax() + 1
    labels[labels == largest_label_id] = 0

    return labels != 0

def add_cap(I, cap_thickness, cap_val):
    """Add caps to 3D image.
    Caps are added on both ends along the Z-direction (first dataset dimension). The thickness and color (Grey Value) of the added caps can be specified.

    Parameters
    ----------
    I
        3D data. Zeroes as background.
    cap_thickness : int
        Cap thickness in pixels.
    cap_val : float
        Cap grey value.

    Returns
    ----------
    I_cap
        Image with caps added.
    """

    I_cap = np.ones([I.shape[0]+2*cap_thickness, I.shape[1], I.shape[2]], I.dtype)*cap_val
    I_cap[cap_thickness:-cap_thickness, :, :] = I
    return I_cap

def embed(I, embed_depth, embed_dir, embed_val=None, pad=0, makecopy=False):
    """Add embedding to 3D image.
    Direction and depth of the embedded region should be given. Zeroes in the input image is considered to be background.

    Parameters
    ----------
    I
        3D data. Zeroes as background.
    embed_depth : int
        Embedding depth in pixels.
    embed_dir : str
        Embedding direction. Can be "-x", "+x", "-y", "+y", "-z", or "+z".
    embed_val : float
        Embedding grey value.
    pad = int
        Padding around bounding box of embedded area.
    makecopy : bool
        Make copy of the input image.

    Returns
    ----------
    I
        Embedded image. Same size as the input one.
    BW_embedding
        BW mask of the embedding area.
    """

    if embed_val is None:
        embed_val = I.max() + 1

    # binarize the input image
    BW_I = np.zeros(I.shape, dtype='bool')
    BW_I[I>0] = True

    # init embedding mask
    BW_embedding = np.zeros(BW_I.shape, dtype='bool')

    if embed_dir == "-z":
        dir = -1

        # start the embedding at first non-zero voxel
        embed_start = np.where(np.max(BW_I.max(1), 1) == True)[-1][-1]

        # project embedded area and find size of embedding
        bbox_origin, bbox_size = ru.bbox(BW_I[embed_start + (dir * embed_depth):, :, :], pad)

        # create embedding mask
        BW_embedding[embed_start + (dir * embed_depth):, bbox_origin[0]:bbox_origin[0] + bbox_size[0], bbox_origin[1]:bbox_origin[1] + bbox_size[1]] = True

    elif embed_dir == "+z":
        dir = 1

        # start the embedding at first non-zero voxel
        embed_start = np.where(np.max(BW_I.max(1), 1) == True)[0][0]

        # project embedded area and find size of embedding
        bbox_origin, bbox_size = ru.bbox(BW_I[:embed_start + (dir * embed_depth), :, :], pad)

        # create embedding mask
        BW_embedding[:embed_start + (dir * embed_depth), bbox_origin[0]:bbox_origin[0] + bbox_size[0], bbox_origin[1]:bbox_origin[1] + bbox_size[1]] = True

    elif embed_dir == "-x":
        dir = -1

        # start the embedding at first non-zero voxel
        embed_start = np.where(np.max(BW_I.max(0), 0) == True)[-1][-1]

        # project embedded area and find size of embedding
        bbox_origin, bbox_size = ru.bbox(BW_I[:, :, embed_start + (dir * embed_depth):], pad)

        # create embedding mask
        BW_embedding[bbox_origin[2]:bbox_origin[2] + bbox_size[2], bbox_origin[0]:bbox_origin[0] + bbox_size[0], embed_start + (dir * embed_depth):] = True

    elif embed_dir == "+x":
        dir = +1

        # start the embedding at first non-zero voxel
        embed_start = np.where(np.max(BW_I.max(0), 0) == True)[0][0]

        # project embedded area and find size of embedding
        bbox_origin, bbox_size = ru.bbox(BW_I[:, :, :embed_start + (dir * embed_depth)], pad)

        # create embedding mask
        BW_embedding[bbox_origin[2]:bbox_origin[2] + bbox_size[2], bbox_origin[0]:bbox_origin[0] + bbox_size[0], :embed_start + (dir * embed_depth)] = True

    elif embed_dir == "+y":
        dir = -1

        # start the embedding at first non-zero voxel
        embed_start = np.where(np.max(BW_I.max(0), 1) == True)[-1][-1]

        # project embedded area and find size of embedding
        bbox_origin, bbox_size = ru.bbox(BW_I[:, embed_start + (dir * embed_depth):, :], pad)

        # create embedding mask
        BW_embedding[bbox_origin[2]:bbox_origin[2] + bbox_size[2], embed_start + (dir * embed_depth):, bbox_origin[1]:bbox_origin[1] + bbox_size[1]] = True

    elif embed_dir == "-y":
        dir = +1

        # start the embedding at first non-zero voxel
        embed_start = np.where(np.max(BW_I.max(0), 1) == True)[0][0]

        # project embedded area and find size of embedding
        bbox_origin, bbox_size = ru.bbox(BW_I[:, :embed_start + (dir * embed_depth), :])

        # create embedding mask
        BW_embedding[bbox_origin[2]:bbox_origin[2] + bbox_size[2], :embed_start + (dir * embed_depth), bbox_origin[1]:bbox_origin[1] + bbox_size[1]] = True

    else:
        raise IOError("EMBED_DIR parameter unknown. Valid entries are -x, +x, -y, +y, -z, and +z.")

    # emboss embedding mask with the masked input image
    BW_embedding = remove_unconnected(BW_embedding & ~BW_I)

    # assign embedding val to input image
    if makecopy:
        I_output = I.copy()
        I_output[BW_embedding] = embed_val
        return I_output, BW_embedding
    else:
        I[BW_embedding] = embed_val
        return I, BW_embedding

def periosteummask(bwimage, closepixels=10, closevoxels=0, remove_objects_smaller_than=None, removeunconn=True, verbose=False):
    """Binary mask of periosteum (whole bone).

    Parameters
    ----------
    bwimage : bool
        Binary image. Can be 2D or 3D.
    closepixels : int
        Radius of DISK structuring element for 2D image closing.
    closevoxels : int
        Radius of CUBE structuring element for final 3D image closing.
    remove_objects_smaller_than : int
        Remove objects smaller than given size before periosteum mask calculation.
    removeunconn : bool
        Remove unconnected clusters of pixels/voxels from the calculated mask.
    verbose : bool
        Activate verbose output.

    Returns
    -------
    perimask : bool
        Binary mask of the whole bone (periosteum mask).
    """

    # verbose output
    if verbose:
        logging.basicConfig(level=logging.INFO)

    if remove_objects_smaller_than:
        logging.info('Preliminary removal of objects smaller than {} pixels.'.format(remove_objects_smaller_than))
        bwimage = morphology.remove_small_objects(bwimage, min_size=remove_objects_smaller_than)

    if bwimage.ndim == 3:

        # output arrays initialization
        perimask = np.zeros(np.shape(bwimage), dtype=bool)

        # 2D slice-wise imclose and fill
        logging.info('2D slice-wise image closing and filling.\n Structuring element DISK of radius: {}'.format(closepixels))

        for slice in tqdm(range(0, bwimage.shape[0])):
            perimask[slice,:,:] = ndimage.binary_fill_holes(morphology.binary_closing(bwimage[slice,:,:], morphology.disk(closepixels)))

        if removeunconn:
            # remove isolated clusters
            logging.info("Removing isolated clusters of voxels.")
            perimask = remove_unconnected(perimask)

        if closevoxels > 0:
            # final 3D imclose
            logging.info('Final 3D image closing.\n Structuring element CUBE of radius: {}'.format(closepixels))
            perimask = morphology.binary_closing(perimask, morphology.cube(closevoxels))

    elif bwimage.ndim == 2:
        # imclose and fill
        perimask = ndimage.binary_fill_holes(morphology.binary_closing(bwimage, morphology.disk(closepixels)))

        # remove isolated clusters
        if removeunconn:
            # remove isolated clusters
            logging.info("Removing isolated clusters of voxels.")
            perimask = remove_unconnected(perimask)

    return perimask